from pathlib import Path

import libsbml
import requests
import sympy2jax
from sbmlmath import SBMLMathMLParser

from enzax.kinetic_model import KineticModelSbml


def _get_libsbml_model_from_doc(doc):
    if doc.getModel() is None:
        raise ValueError("Failed to load the SBML model")
    elif doc.getModel().getNumFunctionDefinitions():
        convert_config = (
            libsbml.SBMLFunctionDefinitionConverter().getDefaultProperties()
        )
        doc.convert(convert_config)
    model = doc.getModel()
    return model


def load_libsbml_model_from_file(file_path: Path) -> libsbml.Model:
    """Load a libsbml.Model object from local file at file_path.

    Args:
        file_path: The path to the SBML file.

    Returns:
        A libsbml.Model object.

    """
    reader = libsbml.SBMLReader()
    doc = reader.readSBML(file_path)
    return _get_libsbml_model_from_doc(doc)


def load_libsbml_model_from_url(url: str) -> libsbml.Model:
    """Load a libsbml.Model object from a url.

    Args:
        url: The url to the SBML file.

    Returns:
        A libsbml.Model object

    """
    reader = libsbml.SBMLReader()
    with requests.get(url) as response:
        doc = reader.readSBMLFromString(response.text)
    return _get_libsbml_model_from_doc(doc)


def sbml_to_sympy(model):
    reactions_sbml = model.getListOfReactions()
    reactions_sympy = [
        (
            SBMLMathMLParser().parse_str(
                libsbml.writeMathMLToString(
                    libsbml.parseL3Formula(
                        libsbml.formulaToL3String(r.getKineticLaw().getMath())
                    )
                )
            )
        )
        for r in reactions_sbml
    ]
    return reactions_sympy


def sympy_to_enzax(reactions_sympy):
    sym_module = sympy2jax.SymbolicModule(reactions_sympy)
    return sym_module


def get_sbml_parameters(model: libsbml.Model) -> dict:
    local_parameters = {
        p.getId(): p.getValue()
        for r in model.getListOfReactions()
        for p in r.getKineticLaw().getListOfParameters()
    }
    compartment_volumes = {
        c.getId(): c.volume for c in model.getListOfCompartments()
    }
    unbalanced_species = {
        u.getId(): u.getInitialConcentration()
        for u in model.getListOfSpecies()
        if u.boundary_condition
    }
    global_parameters = {
        p.getId(): p.getValue()
        for p in model.getListOfParameters()
        if p.constant
    }
    return {
        **local_parameters,
        **compartment_volumes,
        **unbalanced_species,
        **global_parameters,
    }


def get_reaction_stoichiometry(reaction: libsbml.Reaction) -> dict[str, float]:
    reactants = reaction.getListOfReactants()
    products = reaction.getListOfProducts()
    reactant_stoichiometries, product_stoichiometries = (
        {s.getSpecies(): coeff * s.getStoichiometry() for s in list_of_species}
        for list_of_species, coeff in [(reactants, -1.0), (products, 1.0)]
    )
    return reactant_stoichiometries | product_stoichiometries


def get_kinetic_model_from_sbml(
    libsbml_model: libsbml.Model,
) -> KineticModelSbml:
    """Turn a libsbml.Model into a KineticModelSbml.

    Args:
        libsbml_model: The libsbml.Model to convert.

    Returns:
        A KineticModelSbml

    """
    species = [s.getId() for s in libsbml_model.getListOfSpecies()]
    balanced_species = [
        b.getId()
        for b in libsbml_model.getListOfSpecies()
        if not b.boundary_condition
    ]
    reactions = [
        reaction.getId() for reaction in libsbml_model.getListOfReactions()
    ]
    stoichiometry = {
        reaction.getId(): get_reaction_stoichiometry(reaction)
        for reaction in libsbml_model.getListOfReactions()
    }
    sym_module = get_sbml_sym_module(libsbml_model)
    return KineticModelSbml(
        stoichiometry=stoichiometry,
        species=species,
        reactions=reactions,
        balanced_species=balanced_species,
        sym_module=sym_module,
    )


def get_sbml_sym_module(model: libsbml.Model):
    reactions_sympy = sbml_to_sympy(model)
    return sympy_to_enzax(reactions_sympy)


def sbml_to_enzax(
    libsbml_model: libsbml.Model,
) -> tuple[KineticModelSbml, dict]:
    """Turn a libsbml.Model into a KineticModelSbml plus parameters.

    Args:
        libsbml_model: The libsbml.Model to convert.

    Returns:
        A tuple of a KineticModelSbml and a dictionary of parameters

    """
    parameters = get_sbml_parameters(libsbml_model)
    model = get_kinetic_model_from_sbml(libsbml_model)
    return model, parameters



from pathlib import Path

import jax
import jax.numpy as jnp

from enzax.sbml import load_libsbml_model_from_file, sbml_to_enzax
from enzax.steady_state import get_steady_state

jax.config.update("jax_enable_x64", True)

SBML_FILE_PATH = Path("tests") / "data" / "exampleode.xml"


def main():
    model_libsbml = load_libsbml_model_from_file(SBML_FILE_PATH)
    model, parameters = sbml_to_enzax(model_libsbml)

    y0 = jnp.array([2.0, 4])
    print(f"Flux at {str(y0)}: " + str(model.flux(y0, parameters)))

    print(
        f"dcdt at {str(y0)}, t=1: "
        + str(model.dcdt(conc=y0, parameters=parameters))
    )

    guess = jnp.full((2), 0.01)
    steady_state = get_steady_state(model, guess, parameters)
    print("Steady state: " + str(steady_state))


if __name__ == "__main__":
    main()



from jax import numpy as jnp
import pytest
import importlib.resources
from tests import data
from enzax import sbml
from enzax.steady_state import get_steady_state

exampleode_file = importlib.resources.files(data) / "exampleode.xml"


@pytest.mark.parametrize(
    "file_path",
    [
        exampleode_file,
    ],
)
def test_load_libsbml_model(file_path):
    sbml.load_libsbml_model_from_file(file_path)


@pytest.mark.parametrize(
    ["path", "expected", "guess"],
    [
        (
            exampleode_file,
            jnp.array([0.3230166, 3.02209784]),
            jnp.array([0.01, 0.01]),
        ),
    ],
)
def test_sbml_to_enzax(path, expected, guess):
    libsbml_model = sbml.load_libsbml_model_from_file(path)
    model, parameters = sbml.sbml_to_enzax(libsbml_model)
    steady_state = get_steady_state(model, guess, parameters)
    assert jnp.isclose(steady_state, expected).all()



"""Demonstration of how to find a steady state and its gradients with enzax."""

import time

import jax
from jax import numpy as jnp

from enzax.examples import linear as example
from enzax.steady_state import get_steady_state

BAD_GUESS = jnp.full(example.steady_state.shape, 0.01)
GOOD_GUESS = example.steady_state


def main():
    """Function for testing the steady state solver."""
    model = example.model
    parameters = example.parameters

    # compare good and bad guess
    for guess in [BAD_GUESS, GOOD_GUESS]:
        # solve once for jitting
        steady = get_steady_state(model, GOOD_GUESS, parameters)
        jax.jacrev(model.dcdt, argnums=1)(steady, parameters)
        # timer on
        start = time.time()
        conc_steady = get_steady_state(model, guess, parameters)
        jac = jax.jacrev(model.dcdt, argnums=1)(conc_steady, parameters)
        # timer off
        runtime = (time.time() - start) * 1e3
        sv = model.dcdt(conc_steady, parameters)
        flux = model.flux(conc_steady, parameters)
        print(f"Results with starting guess {guess}:")
        print(f"\tRun time in milliseconds: {round(runtime, 4)}")
        print(f"\tSteady state concentration: {conc_steady}")
        print(f"\tFlux: {flux}")
        print(f"\tSv: {sv}")
        print(f"\tLog substrate Km Jacobian: {jac['log_substrate_km']}")
        print(f"\tDgf Jacobian: {jac['dgf']}")


if __name__ == "__main__":
    main()



"""Module containing enzax's definition of a kinetic model."""

from abc import abstractmethod
from typing import Any

import equinox as eqx
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array, Float, PyTree
from numpy.typing import NDArray

from enzax.rate_equation import RateEquation


def get_ix_from_list(s: str, list_of_strings: list[str]):
    return next(i for i, si in enumerate(list_of_strings) if si == s)


class KineticModel(eqx.Module):
    """Structural information about a kinetic model."""

    stoichiometry: dict[str, dict[str, float]] = eqx.field(static=True)
    species: list[str] = eqx.field(static=True)
    reactions: list[str] = eqx.field(static=True)
    balanced_species: list[str] = eqx.field(static=True)
    unbalanced_species: list[str] = eqx.field(static=True, init=False)
    species_to_dgf_ix: NDArray[np.int16] = eqx.field(
        static=True, default=slice(None)
    )
    balanced_species_ix: NDArray[np.int16] = eqx.field(static=True, init=False)
    unbalanced_species_ix: NDArray[np.int16] = eqx.field(
        static=True, init=False
    )
    S: NDArray[np.float64] = eqx.field(static=True, init=False)

    def __post_init__(self, species_to_dgf_ix=None):
        self.unbalanced_species = [
            s for s in self.species if s not in self.balanced_species
        ]
        self.balanced_species_ix = np.array(
            [get_ix_from_list(s, self.species) for s in self.balanced_species],
            dtype=np.int16,
        )
        self.unbalanced_species_ix = np.array(
            [
                get_ix_from_list(s, self.species)
                for s in self.unbalanced_species
            ],
            dtype=np.int16,
        )
        S = np.zeros(shape=(len(self.species), len(self.reactions)))
        for ix_reaction, reaction in enumerate(self.reactions):
            for species_i, coeff in self.stoichiometry[reaction].items():
                ix_species = get_ix_from_list(species_i, self.species)
                S[ix_species, ix_reaction] = coeff
        self.S = S.astype(np.float64)

    def tree_flatten(self):
        children = (
            self.stoichiometry,
            self.species,
            self.reactions,
            self.balanced_species,
            self.species_to_dgf_ix,
        )
        aux_data = None
        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        return cls(*children)

    def get_conc(self, balanced, log_unbalanced):
        conc = jnp.zeros(self.S.shape[0])
        conc = conc.at[self.balanced_species_ix].set(balanced)
        conc = conc.at[self.unbalanced_species_ix].set(jnp.exp(log_unbalanced))
        return conc

    @abstractmethod
    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
        parameters: PyTree,
    ) -> Float[Array, " n"]: ...

    def dcdt(
        self,
        conc: Float[Array, " n_balanced"],
        parameters: PyTree,
    ) -> Float[Array, " n_balanced"]:
        """Get the rate of change of balanced species concentrations.

        :param conc: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        :param parameters: A PyTree of parameters.

        :return: a one dimensional array of floats representing the rate of change of balanced species concentrations. Has same size as self.structure.ix_balanced.
        """  # Noqa: E501
        v = self.flux(conc, parameters)
        sv = self.S @ v
        return jnp.array(sv[self.balanced_species_ix])

    def __call__(self, t, y, parameters):
        return self.dcdt(y, parameters)


class RateEquationModel(KineticModel):
    """A kinetic model that specifies its fluxes using RateEquation objects."""

    rate_equations: list[RateEquation] = eqx.field(
        static=True, default_factory=list
    )

    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
        parameters: PyTree,
    ) -> Float[Array, " n"]:
        """Get fluxes from balanced species concentrations.

        :param conc_balanced: a one dimensional array of positive floats representing concentrations of balanced species. Must have same size as self.structure.ix_balanced

        :return: a one dimensional array of (possibly negative) floats representing reaction fluxes. Has same size as number of columns of self.structure.S.

        """  # Noqa: E501
        conc = self.get_conc(conc_balanced, parameters["log_conc_unbalanced"])
        flux_list = []
        for reaction_ix, (reaction_id, rate_equation) in enumerate(
            zip(self.reactions, self.rate_equations)
        ):
            ipt = rate_equation.get_input(
                parameters=parameters,
                reaction_id=reaction_id,
                reaction_stoichiometry=self.S[:, reaction_ix],
                species_to_dgf_ix=self.species_to_dgf_ix,
            )
            flux_list.append(rate_equation(conc, ipt))
        return jnp.array(flux_list)


class KineticModelSbml(KineticModel):
    sym_module: Any = eqx.field(static=True, default=None)

    def flux(
        self,
        conc_balanced: Float[Array, " n_balanced"],
        parameters,
    ) -> Float[Array, " n"]:
        flux = jnp.array(
            self.sym_module(
                **parameters,
                **dict(zip(self.balanced_species, conc_balanced)),
            )
        )
        return flux
